/* wbxml.cpp --- wml tokenizer */

/*
 * Copyright (C) 2000 Roger Willcocks
 * rogerw@centipede.co.uk
 *
 * 06 feb 00	rkww	initial release
 * 07 feb 00	rkww	fixed minor coding errors
 * 08 feb 00	rkww	created single parser class
 * 15 feb 00	rkww	turned back into C ...
 * 19 feb 00	rkww	more comments
 * 20 feb 00	rkww	removed test code
 * 07 mar 00	rkww	made string table 8-bit clean
 * 23 jun 00	rkww	emit wmlc error message
 */

/*
 * This program reads wml source and tokenizes to wmlc format. It
 * makes no attempt to ensure that tags are properly nested or that
 * attributes are properly specified. It catches most syntactic
 * errors though.
 *
 * For 'lint'ing wml I suggest http://www.ltg.ed.ac.uk/~richard/rxp.html
 */

/*
 * This code is thread-safe. The first instance of WmlParser initializes
 * internal data structures, these are destroyed when the last instance
 * is deleted.
 */

#include "wbxml.h"

#ifdef __cplusplus /* this module is actually 'C' code */
extern "C" {
#endif

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>

#ifdef _WIN32
#define strcasecmp stricmp
#endif


/* symbol types / flags */

#define ATTRNAME 0x1000
#define COMPLEXATTR 0x2000
#define BOOLEANATTR 0x4000
#define TAGNAME 0x8000

struct _literal;
struct _token;

int allocs;
void* xmalloc(int t)
{
    allocs++;
    return malloc(t);
}

void xfree(void *p)
{
    allocs--;
    free(p);
}

#define malloc xmalloc
#define free   xfree

typedef struct _wmlparser
{
    int linenumber; /* for error handling */

    char* outputbuffer;
    int outputptr;  /* pointer into outputbuffer */

    struct _literal* stringtable; /* head of string table list */
    struct _literal* hook; /* temporary used during tokenization */

    jmp_buf aaagh;

    struct _symbol *userdictionary; /* user string table */

} wmlparser;

#define MAXAMPLEN       20
#define MAXATTRNAMELEN 200
#define MAXATTRVALLEN  200
#define MAXTAGNAMELEN  200
#define MAXCONTENTLEN 2000


/* format of a string dictionary entry */

typedef struct _symbol
{
    struct _symbol *left, *right;	/* binary tree implementation */
    char *text;		/* the string */
    int wiredval;	/* intern value if system name */
    int subst;		/* flags (type, wml tag value) */
    int dupe;		/* reference count for this string */
    int stringoffset;	/* position of this string in string table (if dupe) */
} symbol;


#define QUOTE(sym) #sym
#define JOIN3(x, y, z) x##y##z
#define Y(sym,info) JOIN3(_,sym,_),


/* build magic _symbolname_ constants corresponding to the entries in
   the tagmap table (these are the 'wiredval' intern values for
   internal symbols) */

enum wiredsymbolname {
    _UnusedZero_ = 0,
#include "wmltags.h"
    _accept_charset_,
    _http_equiv_,
    _xml_lang_,
    WiredSymbolCount
};

#undef Y


typedef struct _tagmapentry {
    char* wired;
    int subst;
} tagmapentry;


#define Y(sym,info) QUOTE(sym), info,

/* construct a map of well-known names to tag values, corresponding
   1:1 to the wiredsymbol table above. The strings are interned to
   symbol* at boot time. */

tagmapentry tagmap[] = {
    0, 0,
#include "wmltags.h"
    "accept-charset", 0x5 | ATTRNAME,
    "http-equiv", 0x5a | COMPLEXATTR | ATTRNAME,
    "xml:lang",   0x50 | ATTRNAME
};

#undef Y

static int refcount = 0; /* number of parsers currently instantiated (for multi-thread) */


/* single shared system dictionary read-only to all but the first
   instance, containing all the internal symbols */

static symbol *systemdictionary;


/* intern a string */

static char* strdupe(char *p)
{
    char *q = (char*)malloc(strlen(p) + 1);
    strcpy(q, p);
    return q;
}


/* end of the game */

/* <wml><card><p>error text</p></card></wml> */
static char preamble[] = { 0x01, 0x04, 0x6a, 0x00, 0x7f, 0x67, 0x60, 0x03 };
static char postamble[] = { 0x00, 0x01, 0x01, 0x01 };

static void Error(wmlparser* _this, char* fmt, ...)
{
    va_list marker;
    int len;

    memcpy(_this->outputbuffer, preamble, sizeof(preamble));
	 
    va_start(marker, fmt);
    vsprintf(_this->outputbuffer + sizeof(preamble), fmt, marker);
    va_end(marker);

    len = sizeof(preamble) + strlen(_this->outputbuffer + sizeof(preamble));
    sprintf(_this->outputbuffer + len, " on line %d", _this->linenumber);
    len += strlen(_this->outputbuffer + len);

    memcpy(_this->outputbuffer + len, postamble, sizeof(postamble));

    _this->outputptr = len + sizeof(postamble); /* previously returned -1 to signal failure */
    longjmp(_this->aaagh, 1);
}


static symbol* NewSymbol(char* _text)
{
    symbol* _this = (symbol*)malloc(sizeof(symbol));

    _this->left = _this->right = 0;
    _this->text = strdupe(_text);
    _this->stringoffset = 0;
    _this->wiredval = 0;
    _this->subst = 0;
    _this->dupe = 0;

    return _this;
}


static void DeleteSymbol(symbol* _this)
{
    if (_this->left)
	DeleteSymbol(_this->left);

    if (_this->right)
	DeleteSymbol(_this->right);

    free(_this->text);
    free(_this);
}


/*
 * symbol table maintenance, look up a symbol in selected dictionary:
 * optionally create symbol if not found. Uses a binary tree implementation.
 */

static symbol* lookup(wmlparser* _this, char* what, int create, int isstring)
{
    symbol **vp = (isstring) ? &_this->userdictionary : &systemdictionary;

    int t;
    for (;;) {
	if (*vp == 0) {
	    if (create)
		*vp = NewSymbol(what);

	    return *vp;
	}
	if ((t = strcmp((*vp)->text, what)) == 0)
	    return *vp;

	else 
	    vp = (t < 0) ? & (*vp)->left : & (*vp)->right;
    }
}


/*
 * emit a single octet
 */
static void emit(wmlparser* _this, int octet)
{
    _this->outputbuffer[_this->outputptr++] = octet;
}


/*
 * emit a multibyte int
 */
static void emitint(wmlparser* _this, int mbint, int continuation)
{
    if (mbint >= 0x80)
	emitint(_this, mbint >> 7, 1);
    mbint &= 0x7f;
    if (continuation)
	mbint |= 0x80;
    emit(_this, mbint);
}


/*
 * emit a UTF-8 null-terminated string
 */
static void emitstring(wmlparser* _this, char* buffer)
{
    int c;
    while ((c = (*buffer++ & 0xff)) != 0) {
	if (c & 0x80)
	    emit(_this, 0xc2); /* UTF-8 lead byte */
	emit(_this, c);
    }
    emit(_this, 0x00);
}


/*
 * calculate length of UTF-8 null-terminated string
 */
static int stringlength(char* buffer)
{
    int c;
    int i = 0;
    while ((c = (*buffer++ & 0xff)) != 0) {
	if (c & 0x80)
	    i++;
	i++;
    }
    return i + 1;
}


/*
 * build the wired names used internally. Since the string table is a
 * binary tree, it's a bad idea to feed the strings in order (we'd
 * just end up with a linked list) so insert in binary order.
 */
static void InitWired(wmlparser* _this, int low, int high)
{
    int i = (low + high) / 2;

    symbol* sym = lookup(_this, tagmap[i].wired, 1, 0);
    sym->subst = tagmap[i].subst;
    sym->wiredval = i;

    if (low < i)
	InitWired(_this, low, i);

    if (i + 1 < high)
	InitWired(_this, i + 1, high);
}


/*
 * literals are the internal building blocks of the parser. many literals
 * may refer to the same symbol instance.
 */
typedef struct _literal
{
    symbol* token;	/* the actual text (+ substitution info) */
    int isVariable;	/* this is a $(variable) */
    struct _literal* next; /* string of literals forming an element */
    int escape;		/* this variable's escape type */
    struct _literal* nextstring; /* next string in stringtable */

} literal;


static literal* NewLiteral(symbol* _token)
{
    literal* _this = (literal*)malloc(sizeof(literal));

    _this->token = _token;
    _this->isVariable = 0;
    _this->nextstring = 0;
    _this->escape = 0;
    _this->next = 0;

    return _this;
}

static void DeleteLiteral(literal* _this)
{
    free(_this);
}


/*
 * helper function for Tokenize
 */
static char* SkipLabel(char *text)
{
    if (*text == '_' || *text == ':' ||
	(*text >= 'a' && *text <= 'z') ||
	(*text >= 'A' && *text <= 'Z'))
	;
    else
	return text;

    while (*text == '_' || *text == ':' ||
	   *text == '.' || *text == '-' ||
	   (*text >= 'a' && *text <= 'z') ||
	   (*text >= 'A' && *text <= 'Z') ||
	   (*text >= '0' && *text <= '9'))
	text++;

    return text;
}


static void Squish(wmlparser* _this, char *text, int isAttribute); /* forward */


/*
 * break text string into string of literals and $(variables)
 */
static literal *Tokenize(wmlparser* _this, char *text, int isAttribute)
{
    literal* head;
    symbol* var;
    char t;

    if (*text == 0)
	return 0;

    /* the literal stream is created by appending to the chain tail
       (pointed to by hook); create a temporary chain head */

    head = _this->hook = NewLiteral((symbol*)0);

    for (;;) {
	char *p = text;
	int isVariable = (p[0] == '$' && p[1] != '$');

	if (isVariable) {
	    text = p + 1;
	    if (p[1] == '(') {
		text = p = p + 2;
		while (*p && *p != ')')
		    p++;
		if (! *p)
		    Error(_this, "missing ')'");
		*p++ = 0;
	    }
	    else
		p = SkipLabel(text);
	}
	else {
	  more:
	    while (*p && *p != '$')
		p++;

	    if (p[0] == '$' && p[1] == '$') {
		strcpy(p, p+1); /* only want one $ in output */
		p++;
		goto more;
	    }
	}

	t = *p;
	*p = 0; /* cut at end of literal */

	if (isVariable) {
	    char *u = strchr(text, ':');

	    if (u &&
		strcasecmp(u, ":e") != 0 &&
		strcasecmp(u, ":escape") != 0 &&
		strcasecmp(u, ":u") != 0 &&
		strcasecmp(u, ":unesc") != 0 &&
		strcasecmp(u, ":n") != 0 &&
		strcasecmp(u, ":noesc") != 0)
		Error(_this, "bad symbol escape format");

	    if (u)
		*u = 0; /* cut at ':' */

	    var = lookup(_this, text, 1, 1); /* create base symbol in string dictionary */

	    if (u)
		*u = ':';

	    _this->hook = _this->hook->next = NewLiteral(var);

	    _this->hook->isVariable = 1;
	    _this->hook->escape = (u ? u[1] : 0);

	    /* first instance of this variable? Hook it into the name table */

	    if (var->dupe++ == 0) {
		_this->hook->nextstring = _this->stringtable;
		_this->stringtable = _this->hook;
	    }
	}
	else /* not a variable */
	    Squish(_this, text, isAttribute);

	*p = t;
	text = p;

	if (t == 0) { /* done */
	    literal* lit = head->next; /* discard temporary chain head */
	    DeleteLiteral(head);
	    _this->hook = 0;
	    return lit;
	}
    }
}


typedef struct _attribute
{
    int idiom;			/* single byte representing name */
    symbol* name;		/* the name itself */
    literal* valuelist;		/* Tokenized value */
    char* value_temp;		/* temporary for idiom recognition */
    struct _attribute* next;	/* next attribute for this element */

} attribute;


static int SpotIdioms(attribute* _this); /* forward */

static attribute* NewAttribute(wmlparser* parser, char* _name, char* _value)
{
    attribute* _this = (attribute*)malloc(sizeof(attribute));

#if 1
    _this->name = lookup(parser, _name, 0, 0);
    if (! _this->name || (_this->name->subst & ATTRNAME) == 0)
	Error(parser, "not a valid wml attribute: '%s'", _name);
#else
    _this->name = lookup(parser, _name, 1, 0);
#endif
    _this->value_temp = _value;
    _this->idiom = SpotIdioms(_this); /* updates value_temp */

    /* Tokenize whatever's left, parse for $(variables) */
    _this->valuelist = Tokenize(parser, _this->value_temp, 1);
    _this->value_temp = 0;
    _this->next = 0;

    return _this;
}

static void DeleteAttribute(attribute* _this)
{
    while (_this->valuelist) {
	literal* p = _this->valuelist;
	_this->valuelist = _this->valuelist->next;
	DeleteLiteral(p);
    }
    free(_this);
}


/*
 * simple-minded string-based input stream
 */
typedef struct _channel
{
    int c;		/* one character lookahead */
    wmlparser* parser;	/* context for error handling */
    char *inptr;	/* pointer into input buffer */
} channel;


channel* NewChannel(wmlparser* _parser, char *_inptr)
{
    channel* _this = (channel*)malloc(sizeof(channel));

    _this->inptr = _inptr;
    _this->parser = _parser;
    if ((_this->c = *(_this->inptr)++) == 0) /* prime the pump... */
	_this->c = -1;

    return _this;
}


static void DeleteChannel(channel* _this)
{
    free(_this);
}


static int ispeekwhite(channel* _this) {
    return (_this->c == ' ' || _this->c == '\r' || _this->c == '\n' || _this->c == '\t');
}

static int peekchannel(channel* _this) {
    return _this->c;
}

static int readchannel(channel* _this, int checkcomments)
{
    symbol* tag;
    char charbuff[MAXAMPLEN];
    int i;

    int t = _this->c;

    if (t == -1 || (_this->c = *(_this->inptr)++) == 0)
	_this->c = -1;

    if (t == -1)
	Error(_this->parser, "unexpected end-of-file");

    if (t == '\n')
	_this->parser->linenumber++;

    if (checkcomments && t == '<' && _this->c == '!') { /* weak test */
	(void)readchannel(_this, 0); /* ! */
	if (readchannel(_this, 0) != '-' || readchannel(_this, 0) != '-')
	    Error(_this->parser, "bad comment start");
	for (;;) {
	    if (readchannel(_this, 0) == '-' && peekchannel(_this) == '-') {
		(void)readchannel(_this, 0);
		if (peekchannel(_this) == '>') {
		    (void)readchannel(_this, 0);
		    return readchannel(_this, 1); /* finally, return the next character */
		}
	    }
	}
    }

    if (t != '&')
	return t;

    /* process &amp; etc. */

    i = 0;
    while (peekchannel(_this) != ';') {
	charbuff[i++] = readchannel(_this, 1);
	if (i == sizeof(charbuff))
	    Error(_this->parser, "constant too long");
    }
    readchannel(_this, 1);
    charbuff[i++] = 0;

    if (charbuff[0] == '#') {
	t = 0;
	if (charbuff[1] == 'x')
	    sscanf(&charbuff[2], "%x", &t);
	else
	    sscanf(&charbuff[1], "%d", &t);
	if (! t)
	    Error(_this->parser, "bad constant");
	return t | 0x8000;
    }

    tag = lookup(_this->parser, charbuff, 0, 0);

    switch (tag->wiredval) {
      case _quot_: t = '"'; break;
      case _amp_:  t = '&'; break;
      case _apos_: t = '\''; break;
      case _lt_:   t = '<'; break;
      case _gt_:   t = '>'; break;
      case _nbsp_: t = 160; break;
      case _shy_:  t = 173; break;
      default:
	Error(_this->parser, "bad character name");
    }

    return t | 0x8000; /* don't match literal characters during parse */
}


void skipwhite(channel* _this)
{
    while (ispeekwhite(_this))
	readchannel(_this, 1);
}


typedef struct _element
{
    symbol* name;		/* name of this element */
    struct _element* next;	/* sibling element of this element */

    /* an element either contains literal content or a child element */

    literal* contentlist;	/* linked list of content and $(variables) */
    				/* or */
    attribute* attrlist;	/* linked list of attributes */
    struct _element* memberlist;/* child elements of this element */

} element;


static element* NewElement()
{
    element* _this = (element*)malloc(sizeof(element));

    _this->name = 0;
    _this->attrlist = 0;
    _this->contentlist = 0;
    _this->memberlist = 0;
    _this->next = 0;

    return _this;
}


static void DeleteElement(element* _this)
{
    while (_this->contentlist) {
	literal* p = _this->contentlist;
	_this->contentlist = _this->contentlist->next;
	DeleteLiteral(p);
    }
    while (_this->attrlist) {
	attribute* p = _this->attrlist;
	_this->attrlist = _this->attrlist->next;
	DeleteAttribute(p);
    }
    while (_this->memberlist) {
	element* p = _this->memberlist;
	_this->memberlist = _this->memberlist->next;
	DeleteElement(p);
    }
    free(_this);
}


/* we have just read past an '<' and created 'this' element */
/* <^tag */

static void LoadElement(element* _this, wmlparser* parser, channel *chan)
{
    attribute *lastattr = 0;
    element *lastmember = 0;

    attribute* ap;
    element* ep;

    char tagname[MAXTAGNAMELEN];
    char namebuff[MAXATTRNAMELEN];
    char valuebuff[MAXATTRVALLEN];

    int i, match;

    int opentagline = parser->linenumber;

    i = 0;
    while (! ispeekwhite(chan) && peekchannel(chan) != '/' && peekchannel(chan) != '>') {
	tagname[i++] = readchannel(chan, 1);
	if (i == sizeof(tagname))
	    Error(parser, "tag name too long");
    }
    tagname[i++] = 0;

#if 1
    _this->name = lookup(parser, tagname, 0, 0); /* don't create spurious non-wml tags */
    if (! _this->name || (_this->name->subst & TAGNAME) == 0)
	Error(parser, "not a valid wml tag: '%s'", tagname);
#else
    /* careful, this writes to the (shared) systemdictionary ... */
    _this->name = lookup(parser, tagname, 1, 0);
#endif

    skipwhite(chan);

    while (peekchannel(chan) != '>') {

	if (peekchannel(chan) == '/') {
	    (void)readchannel(chan, 1);
	    if (peekchannel(chan) != '>')
		Error(parser, "'>' expected after '/'");
	    (void)readchannel(chan, 1);
	    return;
	}

	/* <tag^ */
	/* <tag attr="val"^ */

	skipwhite(chan);

	i = 0;
	while (peekchannel(chan) != '=') {
	    namebuff[i++] = readchannel(chan, 1);
	    if (i == sizeof(namebuff))
		Error(parser, "attribute name too long");
	}
	namebuff[i++] = 0;

	match = 0;
	(void)readchannel(chan, 1);
	if (peekchannel(chan) == '"' || peekchannel(chan) == '\'')
	    match = readchannel(chan, 1);
	else
	    Error(parser, "'\"' or '\\'' expected after '=' in attribute");

	i = 0;
	while (peekchannel(chan) != match) {
	    valuebuff[i++] = readchannel(chan, 1);
	    if (i == sizeof(valuebuff))
		Error(parser, "attribute value too long");
	}
	(void)readchannel(chan, 1);
	valuebuff[i++] = 0;

	/* create attribute / value pair (with idiom recognition, $(variable) handling) */

	ap = NewAttribute(parser, namebuff, valuebuff);

	/* add attribute name/value pair to the list of attributes for this node */

	if (! _this->attrlist)
	    _this->attrlist = lastattr = ap;
	else
	    lastattr->next = ap;
	lastattr = ap;

	skipwhite(chan);

    } /* collect attrs */


    /* <tag x="y"  ^> */

    (void)readchannel(chan, 1);

    /* <tag x="y"  >^ */

    /* loop collecting members until we see 'our' end tag */

    for (;;) {
	char content[MAXCONTENTLEN];
	i = 0;

	do {
	    if (ispeekwhite(chan)) {
		skipwhite(chan);
		content[i++] = ' '; /* fold runs of spaces */
	    }
	    else {
		content[i++] = readchannel(chan, 1);
	    }
	    if (i == sizeof(content))
		Error(parser, "content too long");

	} while (content[i-1] != '<'); /* '<!--' comments complicate things */

	content[i-1] = 0;

	if (content[0] && strcmp(content, " ") != 0) {

	    /* create a new content node and hook it onto list */

	    element *ep = NewElement();
	    ep->contentlist = Tokenize(parser, content, 0); /* look for $(vars) */

	    if (! _this->memberlist)
		_this->memberlist = lastmember = ep;
	    else
		lastmember->next = ep;
	    lastmember = ep;
	}

	/* <^ */

	if (peekchannel(chan) == '/') { /* </ */
	    int i = 0;
	    (void)readchannel(chan, 1);
	    while (! ispeekwhite(chan) && peekchannel(chan) != '>') {
		tagname[i++] = readchannel(chan, 1);
		if (i == sizeof(tagname))
		    Error(parser, "close tag name too long");
	    }
	    skipwhite(chan);
	    if (readchannel(chan, 1) != '>')
		Error(parser, "missing '>' for close tag");
	    tagname[i++] = 0;
	    if (lookup(parser, tagname, 0, 0) != _this->name)
		Error(parser, "mismatched close tag: '%s' on line %d != '%s'",
		      _this->name->text, opentagline, tagname);
	    return;
	}

	/* create new element and add it to the list of elements for this node,
	   then call ourselves recursively to fill in the details. */

	ep = NewElement();

	if (! _this->memberlist)
	    _this->memberlist = lastmember = ep;
	else
	    lastmember->next = ep;
	lastmember = ep;

	/* <^ */

	LoadElement(ep, parser, chan);

    } /* forever */
}


static void DumpLiteral(literal* _this, wmlparser* parser, int isAttribute, int doescape)
{
    int symboltype;

    if (_this->token->subst) {
	emit(parser, _this->token->subst);
	return;
    }

    symboltype = ((_this->token->dupe > 1) ? 0x80 : 0x40);

    if (_this->isVariable) {
	if (_this->escape == 'e' || _this->escape == 'E')
	    symboltype += 0;
	else if (_this->escape == 'u' || _this->escape == 'U')
	    symboltype += 1;
	else if (_this->escape == 'n' || _this->escape == 'N')
	    symboltype += 2;
	else
	    symboltype += ((doescape) ? 0 : 2);
    }
    else
	symboltype += 3;

    if (symboltype == 0x43)
	symboltype = 3;

    emit(parser, symboltype);

    if (_this->token->dupe > 1)
	emitint(parser, _this->token->stringoffset, 0); /* printf("$(%s)", _this->token->text); */
    else {
	char *p = strchr(_this->token->text, ':');
	if (_this->isVariable && p)
	    *p = 0;
	emitstring(parser, _this->token->text);
	if (p)
	    *p = ':';
    }
}


static void DumpAttribute(attribute* _this, wmlparser* parser)
{
    literal* value = _this->valuelist;

    if ((_this->name->subst & ATTRNAME) == 0) {
#if 1
	Error(parser, "internal error"); /* was checked when attribute was parsed */
#else
	printf(" %s=\"", _this->name->text);
	while (value) {
	    DumpLiteral(value, parser, 1, 0);
	    value = value->next;
	}
	printf("\"");
	return;
#endif
    }

    emit(parser, _this->idiom);

    while (value) {
	DumpLiteral(value, parser, 1,
		    (_this->name->wiredval == _onenterforward_ ||
		     _this->name->wiredval == _onenterbackward_ ||
		     _this->name->wiredval == _href_ ||
		     _this->name->wiredval == _src_)); /* force escaping with certain attributes */

	value = value->next;
    }
}


static void DumpElement(element* _this, wmlparser* parser)
{
    /* it's either content or an element with attributes and members */

    literal* content = _this->contentlist;
    attribute* attr = _this->attrlist;
    element* member = _this->memberlist;

    int subst;

    if (content) {
	while (content) {
	    DumpLiteral(content, parser, 0, 0);
	    content = content->next;
	}
	return;
    }

    if ((_this->name->subst & TAGNAME) == 0) {
#if 1
	Error(parser, "internal error"); /* was already checked when tag was parsed */
#else
	printf("<%s", _this->name->text);

	if (attr) {
	    while (attr) {
		DumpAttribute(attr, p);
		attr = attr->next;
	    }
	}
	if (member) {
	    printf(">");
	    while (member) {
		DumpElement(member, p);
		member = member->next;
	    }
	    printf("</%s>", _this->name->text);
	}
	else
	    printf("/>");

	return;
#endif
    }

    subst = _this->name->subst & 0xff;

    if (attr)
	subst |= 0x80;
    if (member)
	subst |= 0x40;

    emit(parser, subst);

    if (attr) {
	while (attr) {
	    DumpAttribute(attr, parser);
	    attr = attr->next;
	}
	emit(parser, 0x01); /* end of attribute list */
    }
    if (member) {
	while (member) {
	    DumpElement(member, parser);
	    member = member->next;
	}
	emit(parser, 0x01); /* end of member list */
    }
}


/*
 * helper routine for Squish ('scan' and 'Squish' are mutually recursive)
 */
static int scan(wmlparser* _this, char* text, char* match, int subst)
{
    char* p;
    symbol* var;

    if ((p = strstr(text, match)) != 0) {
	*p = 0;

	if (text[0])
	    Squish(_this, text, 1);

	var = lookup(_this, match, 1, 1);
	var->subst = subst;

	_this->hook = _this->hook->next = NewLiteral(var);

	p += strlen(match);

	if (p[0])
	    Squish(_this, p, 1);

	return 1;
    }
    return 0;
}


/*
 * read a string and produce a list of literals (replacing common
 * strings with single byte tokens.) 'Squish' & 'scan' are mutually
 * recursive. Only substitute common strings in attributes, not
 * in contents.
 */
static void Squish(wmlparser* _this, char *text, int isAttribute)
{
    if (isAttribute && strlen(text) >= 3 &&
	(scan(_this, text, ".com/", 0x85) ||
	 scan(_this, text, ".edu/", 0x86) ||
	 scan(_this, text, ".net/", 0x87) ||
	 scan(_this, text, ".org/", 0x88) ||
	 scan(_this, text, "accept", 0x89) ||
	 scan(_this, text, "bottom", 0x8a) ||
	 scan(_this, text, "clear", 0x8b) ||
	 scan(_this, text, "delete", 0x8c) ||
	 scan(_this, text, "help", 0x8d) ||
	 scan(_this, text, "http://www.", 0x8f) ||
	 scan(_this, text, "http://", 0x8e) ||
	 scan(_this, text, "https://www.", 0x91) ||
	 scan(_this, text, "https://", 0x90) ||
	 scan(_this, text, "middle", 0x93) ||
	 scan(_this, text, "nowrap", 0x94) ||
	 scan(_this, text, "onenterbackward", 0x96) ||
	 scan(_this, text, "onenterforward", 0x97) ||
	 scan(_this, text, "onpick", 0x95) ||
	 scan(_this, text, "ontimer", 0x98) ||
	 scan(_this, text, "options", 0x99) ||
	 scan(_this, text, "password", 0x9a) ||
	 scan(_this, text, "reset", 0x9b) ||
	 scan(_this, text, "text", 0x9d) ||
	 scan(_this, text, "top", 0x9e) ||
	 scan(_this, text, "unknown", 0x9f) ||
	 scan(_this, text, "wrap", 0xa0) ||
	 scan(_this, text, "www.", 0xa1)))
	;
    else {
	symbol* var = lookup(_this, text, 1, 1);
	_this->hook = _this->hook->next = NewLiteral(var);

	if (var->dupe++ == 0) {
	    _this->hook->nextstring = _this->stringtable;
	    _this->stringtable = _this->hook;
	}
    }
}


/*
 * helper routine for idiom recognition
 */
static int starts(attribute* _this, char* match, int* subst, int val)
{
    int len = strlen(match);
    if ((int)strlen(_this->value_temp) >= len && strncmp(_this->value_temp, match, len) == 0) {
	*subst = val;
	_this->value_temp += len; /* there's less of this literal now... */
	return 1;
    }
    return 0;
}


/*
 * idiom recognition. Prune attribute name/value pairs. Return the idiom
 * token (if any) adjusting _this->value_temp to remove leading match string.
 */
static int SpotIdioms(attribute* _this)
{
    int s = _this->name->subst & 0xff;

    if (_this->name->subst & BOOLEANATTR) {
	if (_this->name->wiredval == _ordered_) { /* :-( */
	    if (starts(_this, "false", &s, s + 1) ||
		starts(_this, "true", &s, s))
		;
	}
	else {
	    if (starts(_this, "false", &s, s) ||
		starts(_this, "true", &s, s + 1))
		;
	}
    }
    else if (_this->name->subst & COMPLEXATTR) {

	/* should really use tables for this... */

	switch (_this->name->wiredval) {

	  case _align_:
	    if (starts(_this, "bottom", &s, 0x6) ||
		starts(_this, "center", &s, 0x7) ||
		starts(_this, "left", &s, 0x8) ||
		starts(_this, "middle", &s, 0x9) ||
		starts(_this, "right", &s, 0xa) ||
		starts(_this, "top", &s, 0xb))
		;
	    break;

	  case _content_:
	    if (starts(_this, "application/vnd.wap.wmlc;charset=", &s, 0x5c))
		;
	    break;

	  case _href_:
	    if (starts(_this, "http://", &s, 0x4b) ||
		starts(_this, "https://", &s, 0x4c))
		;
	    break;

	  case _http_equiv_:
	    if (starts(_this, "Content-Type", &s, 0x5b) ||
		starts(_this, "Expires", &s, 0x5d))
		;
	    break;

	  case _method_:
	    if (starts(_this, "get", &s, 0x1b) ||
		starts(_this, "post", &s, 0x1c))
		;
	    break;

	  case _mode_:
	    if (starts(_this, "nowrap", &s, 0x1d) ||
		starts(_this, "wrap", &s, 0x1e))
		;
	    break;

	  case _src_:
	    if (starts(_this, "http://", &s, 0x58) ||
		starts(_this, "https://", &s, 0x59))
		;
	    break;

	  case _type_:
	    if (starts(_this, "accept", &s, 0x38) ||
		starts(_this, "delete", &s, 0x39) ||
		starts(_this, "help", &s, 0x3a) ||
		starts(_this, "password", &s, 0x3b) ||
		starts(_this, "onpick", &s, 0x3c) ||
		starts(_this, "onenterbackward", &s, 0x3d) ||
		starts(_this, "onenterforward", &s, 0x3e) ||
		starts(_this, "ontimer", &s, 0x3f) ||
		starts(_this, "options", &s, 0x45) ||
		starts(_this, "prev", &s, 0x46) ||
		starts(_this, "reset", &s, 0x47) ||
		starts(_this, "text", &s, 0x48) ||
		starts(_this, "vnd.", &s, 0x49))
		;
	    break;
	} /* switch */
    }

    return s;
}


static void WritePreamble(wmlparser* _this)
{
    emit(_this, 0x01);       /* wmlparser version 1.1 */
    emitint(_this, 0x04, 0); /* "-//WAPFORUM//DTD WML 1.1//EN" */
    emitint(_this, 0x6a, 0); /* charset=UTF-8 */
}


static void DumpStringTable(wmlparser* _this)
{
    int stringtablelength = 0;
    literal* vp = _this->stringtable;

    while (vp) {
	if (vp->token->dupe > 1) { /* put into string table */
	    vp->token->stringoffset = stringtablelength;
	    stringtablelength += stringlength(vp->token->text);
	}
	vp = vp->nextstring;
    }

    emitint(_this, stringtablelength, 0);

    vp = _this->stringtable;

    while (vp) {
	if (vp->token->dupe > 1)
	    emitstring(_this, vp->token->text);
	vp = vp->nextstring;
    }
}


/*
 * internal top-level routine for parsing wml stream
 */
static void Process(wmlparser* _this, channel* chan)
{
    element *root = NewElement();

    _this->linenumber = 1;
    _this->outputptr = 0;

    if (setjmp(_this->aaagh) == 0) {

	do { /* skip entire DTD definition for now: discard until <w */
	    while (readchannel(chan, 0) != '<')
		;
	} while(peekchannel(chan) != 'w');

	/* <^w */

	/* load the root element, and recursively load all children */
	LoadElement(root, _this, chan);

	if (root->name->wiredval != _wml_)
	    Error(_this, "can't find outer 'wml' tag");

	WritePreamble(_this);
	DumpStringTable(_this);

	DumpElement(root, _this);
    }

    DeleteElement(root);

    if (_this->userdictionary)
	DeleteSymbol(_this->userdictionary);

    _this->userdictionary = 0;
    _this->stringtable = 0;
}


/*------------------------------------------------------------*/

/* user visible implementation */

wmlparser* NewWmlParser()
{
    wmlparser* _this = (wmlparser*)malloc(sizeof(wmlparser));

    _this->userdictionary = 0;
    _this->stringtable = 0;

    /* only write to the systemdictionary at creation of first parser */

    if (refcount++ == 0)
	InitWired(_this, 1, WiredSymbolCount);

    return _this;
}


void DeleteWmlParser(wmlparser* _this)
{
    if (--refcount == 0) {
	DeleteSymbol(systemdictionary);
	systemdictionary = 0;
    }
    free(_this);
}


int ParseWml(wmlparser* _this, char* buffer, char* result)
{
    channel* chan;
    _this->outputbuffer = result;
    chan = NewChannel(_this, buffer);
    Process(_this, chan);
    DeleteChannel(chan);
    return _this->outputptr; /* -1 on error, otherwise length of result */
}

#ifdef __cplusplus
};
#endif

/* end */
